"""Overload provides a class for overloading methods in a class."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from random import randint

from worktoy.parse import maybe
from worktoy.text import monoSpace
from worktoy.meta import Dispatcher, TypeSig

try:
  from typing import Any, Callable, TYPE_CHECKING
except ImportError:
  Any = object
  Callable = object
  TYPE_CHECKING = False

if TYPE_CHECKING:
  FuncList = list[tuple[TypeSig, Callable]]
else:
  FuncList = object


class Overload:
  """Overload provides a class for overloading methods in a class."""

  __function_name__ = None
  __function_owner__ = None
  __func_list__ = None
  __dispatcher_instance__ = None
  __class_method__ = None
  __static_method__ = None
  __instance_id__ = None

  def __init__(self, functionType: type = None) -> None:
    self.__instance_id__ = randint(0, 255)
    self.__static_method__ = False
    self.__class_method__ = False
    self.__func_list__ = []
    if functionType is not None:
      if functionType is staticmethod:
        self.__static_method__ = True
        self.__class_method__ = False
      elif functionType is classmethod:
        self.__static_method__ = False
        self.__class_method__ = True
      else:
        e = """Received invalid function type: '%s'! Only 'staticmethod'
        and 'classmethod' are allowed!"""
        raise TypeError(monoSpace(e % functionType))

  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the field and the owner of the field. When this
    method is called the owner is created, so it is safe for the overload
    instance to create the Dispatcher instance. """
    self.__function_owner__ = owner
    clsName = '%sOverload' % name
    setattr(owner, clsName, self)
    self.__function_name__ = name
    self.__dispatcher_instance__ = Dispatcher(self, self.getFuncList(), )

  def __get__(self, instance: object, owner: type) -> Any:
    """Getter-function"""
    if instance is None:
      return self._classGetter(owner)
    return self._instanceGetter(instance)

  def _classGetter(self, owner: type) -> Any:
    """Getter-function for the class."""
    if self.__class_method__:
      self.__dispatcher_instance__.setBound(owner)
      return self.__dispatcher_instance__
    return self

  def _instanceGetter(self, instance: object) -> Any:
    """Getter-function for the instance. This returns the dispatcher. """
    self.__dispatcher_instance__.setBound(instance)
    return self.__dispatcher_instance__

  def getFuncList(self) -> FuncList:
    """Getter-function for the function dictionary."""
    return maybe(self.__func_list__, [])

  def overload(self, *types) -> Callable:
    """The overload function returns a callable that decorates a function
    with
    the signature. """
    key = TypeSig(*types)

    def decorator(callMeMaybe: Callable) -> Callable:
      """The decorator function that adds the function to the function
      dictionary."""
      if isinstance(callMeMaybe, staticmethod):
        if not self.__static_method__:
          e = """The 'Overload' instance is not a static method!"""
          raise TypeError(e)
      if isinstance(callMeMaybe, classmethod):
        if not self.__class_method__:
          e = """The 'Overload' instance is not a class method!"""
          raise TypeError(e)
      self.__func_list__.append((key, callMeMaybe))
      return callMeMaybe

    return decorator



"""NODEF indicates that each instance owning an AttriBox must receive a
call to its __set__ to explicitly set a value, before it is allowed to
receive __get__ calls. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import Zeroton


class NODEF(Zeroton):
  """NODEF indicates that each instance owning an AttriBox must receive a
  call to its __set__ to explicitly set a value, before it is allowed to
  receive __get__ calls. """



"""DispatchException provides a custom exception raised when an instance
of OverloadDispatcher fails to resolve the correct function from the
given arguments. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import TYPE_CHECKING
except ImportError:
  TYPE_CHECKING = False

if TYPE_CHECKING:
  from worktoy.meta import Overload


class DispatchException(Exception):
  """DispatchException provides a custom exception raised when an instance
  of OverloadDispatcher fails to resolve the correct function from the
  given arguments. """

  __overload_function__ = None
  __pos_args__ = None

  def __init__(self, func: Overload, *args) -> None:
    e = """When calling the overloaded function '%s' the dispatcher could 
    not resolve the correct function from the given arguments: %s!"""
    self.__overload_function__ = func
    self.__pos_args__ = [*args, ]
    argStr = ["""  '%s' of type '%s'""" % (arg, type(arg)) for arg in args]
    argStr = ',\n'.join(argStr)
    errorMsg = e % (func.__function_name__, argStr)
    Exception.__init__(self, errorMsg)

  def getOverload(self) -> Overload:
    """Return the Overload instance."""
    return self.__overload_function__

  def getArgs(self) -> list:
    """Return the arguments causing the exception."""
    return self.__pos_args__



"""OverloadDispatcher class for dispatching overloaded functions."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import DispatchException, TypeSig
from worktoy.text import monoSpace

try:
  from typing import Any, Callable, TYPE_CHECKING
except ImportError:
  Any = object
  Callable = object
  TYPE_CHECKING = False

if TYPE_CHECKING:
  from worktoy.meta import Overload

if TYPE_CHECKING:
  FuncList = list[tuple[TypeSig, Callable]]
else:
  FuncList = object


class Dispatcher:
  """Dispatcher is the class responsible for calling the correct overloaded
  function based on received arguments. """
  __static_method__ = None
  __class_method__ = None
  __func_dict__ = None
  __overload_function__ = None
  __bound_arg__ = None

  def __init__(self, func: Overload, funcList: FuncList) -> None:
    """Initialize the Dispatcher with the function dictionary and the
    function
    type."""
    self.__overload_function__ = func
    self.__func_dict__ = funcList
    self.__static_method__ = func.__static_method__
    self.__class_method__ = func.__class_method__

  def getFuncList(self) -> FuncList:
    """Return the function dictionary."""
    return self.__func_dict__

  def setBound(self, arg: object) -> None:
    """Set the bound argument."""
    self.__bound_arg__ = arg

  def getBound(self, ) -> object:
    """Return the bound argument."""
    return self.__bound_arg__

  def getOverload(self) -> Overload:
    """Return the Overload instance."""
    return self.__overload_function__

  def __call__(self, *args, **kwargs) -> Any:
    """Calling the dispatcher handles the dispatching of the correct
    function"""
    bound = self.getBound()
    argTypes = [type(arg).__name__ for arg in args]
    typeSig, callMeMaybe = None, None
    for (typeSig, callMeMaybe) in self.getFuncList():
      if len(typeSig) != len(args):
        continue
      arg = typeSig.cast(*args)
      if arg is None:
        continue
      break
    else:
      raise DispatchException(self.getOverload(), *args)
    if not callable(callMeMaybe):
      raise DispatchException(self.getOverload(), *args)
    if bound is None:
      return callMeMaybe(*arg, **kwargs)
    return callMeMaybe(bound, *arg, **kwargs)
